#include <windows.h>
#include <stdio.h>
#include <conio.h>
#include <io.h>
#include <fcntl.h>

#define MAX_BUFFERS_WAVIN    400

static int Box ( const char *msg )
{
    //MessageBox ( NULL, msg, "Error Message . . .", MB_OK | MB_ICONEXCLAMATION );
    fprintf ( stderr, "Error: %s\n", msg );
    return -1;
}

static int PrintError ( const char *funcname, MMRESULT res )
{
    char  msg [256];

    switch ( res ) {
    case MMSYSERR_ALLOCATED:   sprintf ( msg, "%s failed:\nresource already allocated (device already open)\n"           , funcname );      break;
    case MMSYSERR_INVALPARAM:  sprintf ( msg, "%s failed:\ninvalid Params\n"                                             , funcname );      break;
    case MMSYSERR_BADDEVICEID: sprintf ( msg, "%s failed:\ndevice identifier out of range\n"                             , funcname );      break;
    case MMSYSERR_NODRIVER:    sprintf ( msg, "%s failed:\nno device driver present (no audio driver)\n"                 , funcname );      break;
    case MMSYSERR_NOMEM:       sprintf ( msg, "%s failed:\nunable to allocate or lock memory (for audio)\n"              , funcname );      break;
    case WAVERR_BADFORMAT:     sprintf ( msg, "%s failed:\nattempted to open with an unsupported waveform-audio format\n", funcname );      break;
    case WAVERR_SYNC:          sprintf ( msg, "%s failed:\ndevice is synchronous but waveOutOpen was\n"                  , funcname );      break;
    default:                   sprintf ( msg, "%s failed:\nunknown error code: 0x%02X\n"                                 , funcname, res ); break;
    case MMSYSERR_NOERROR:     return 0;
    }

    return Box ( msg );
}

typedef struct {
    int      active;
    char*    data;
    size_t   datalen;
    WAVEHDR  hdr;
} oblk_t;

static HWAVEIN       Input_WAVHandle;
//static HWAVEOUT      Output_WAVHandle;
static size_t        BufferBytes;
static WAVEHDR       whi    [MAX_BUFFERS_WAVIN];
static char          *data  [MAX_BUFFERS_WAVIN];
static oblk_t        array  [MAX_BUFFERS_WAVIN];
static unsigned int  NextInputIndex;
static unsigned int  NextOutputIndex;

//////////////////////////////////////////////////////////////////////////////////////////////////////

static int
WindowsAudio_SetupInitInput ( unsigned int  deviceno,
                              unsigned int  SampleFreq,
                              unsigned int  BitsPerSample,
                              unsigned int  Channels )
{
    WAVEFORMATEX  inFormat;
    UINT          deviceID;
    UINT          devices;
    MMRESULT      r;
    UINT          i;

    if ( (devices = waveInGetNumDevs ()) == 0 )
        return Box ( "No audio device present." );

    deviceID = deviceno < devices  ?  deviceno  :  WAVE_MAPPER;

    inFormat.wFormatTag      = WAVE_FORMAT_PCM;
    inFormat.wBitsPerSample  = BitsPerSample;
    inFormat.nChannels       = Channels;
    inFormat.nSamplesPerSec  = SampleFreq;
    inFormat.nBlockAlign     = (inFormat.wBitsPerSample + 7) / 8 * inFormat.nChannels;
    inFormat.nAvgBytesPerSec = inFormat.nSamplesPerSec * inFormat.nBlockAlign;
    inFormat.cbSize          = 0;

    r = waveInOpen ( &Input_WAVHandle, deviceID, &inFormat, 0, 0, CALLBACK_EVENT );
    if ( r != MMSYSERR_NOERROR )
        return PrintError ( "waveInOpen", r );

    BufferBytes = SampleFreq * Channels * ((BitsPerSample + 7) / 8);

    for ( i = 0; i < MAX_BUFFERS_WAVIN; i++ ) {
        whi [i].lpData         = data [i] = malloc (BufferBytes);
        whi [i].dwBufferLength = BufferBytes;
        whi [i].dwFlags        = 0;
        whi [i].dwLoops        = 0;

        r = waveInPrepareHeader ( Input_WAVHandle, whi + i, sizeof (*whi) ); if ( r != MMSYSERR_NOERROR ) return PrintError ( "waveInPrepareHeader", r );
        r = waveInAddBuffer     ( Input_WAVHandle, whi + i, sizeof (*whi) ); if ( r != MMSYSERR_NOERROR ) return PrintError ( "waveInAddBuffer"    , r );
    }
    NextInputIndex = 0;
    waveInStart (Input_WAVHandle);
    return 0;
}

static size_t
WindowsAudio_record ( void *DataPtr )
{
    MMRESULT  r;
    size_t    Bytes;

    if ( whi [NextInputIndex].dwFlags & WHDR_DONE ) {
        Bytes = whi [NextInputIndex].dwBytesRecorded;
        memcpy ( DataPtr, data [NextInputIndex], Bytes );

        r = waveInUnprepareHeader ( Input_WAVHandle, whi + NextInputIndex, sizeof (*whi) ); if ( r != MMSYSERR_NOERROR ) return PrintError ( "waveInUnprepareHeader", r );
        whi [NextInputIndex].lpData         = data [NextInputIndex];
        whi [NextInputIndex].dwBufferLength = BufferBytes;
        whi [NextInputIndex].dwFlags        = 0;
        whi [NextInputIndex].dwLoops        = 0;
        r = waveInPrepareHeader   ( Input_WAVHandle, whi + NextInputIndex, sizeof (*whi) ); if ( r != MMSYSERR_NOERROR ) return PrintError ( "waveInPrepareHeader", r );
        r = waveInAddBuffer       ( Input_WAVHandle, whi + NextInputIndex, sizeof (*whi) ); if ( r != MMSYSERR_NOERROR ) return PrintError ( "waveInAddBuffer"    , r );
        NextInputIndex = (NextInputIndex + 1) % MAX_BUFFERS_WAVIN;
        return  Bytes;
    }

    return 0;
}

static int
WindowsAudio_closeInput ( void )
{
    waveInReset ( Input_WAVHandle );
    waveInClose ( Input_WAVHandle );
    Input_WAVHandle = NULL;

    return 0;
}

void write_wav_header ( int srate, int bps, int nch )
{
    unsigned char Header[44];
    unsigned char *p = Header;
    unsigned int Bytes = (bps+7) / 8;
    unsigned int word32;
    __int64 PCMdataLength = (unsigned int)(1<<31) - (44-8) - 1;

    memcpy ( p, "RIFF", 4 ); p += 4;

    word32 = (unsigned int)(PCMdataLength + (44 - 8));
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);
    *p++ = (unsigned char)(word32 >> 16);
    *p++ = (unsigned char)(word32 >> 24);               // Size of the next chunk

    memcpy ( p, "WAVE", 4 ); p += 4;

    memcpy ( p, "fmt ", 4 ); p += 4;

    *p++ = 0x10; *p++ = 0x00; *p++ = 0x00; *p++ = 0x00; // length of the PCM data declaration = 2+2+4+4+2+2

    //if ( bps < 32 ) {
        *p++ = 0x01; *p++ = 0x00;                       // ACM type 0x0001 = uncompressed linear PCM
    //} else {
    //    *p++ = 0x03; *p++ = 0x00;                       // ACM type 0x0003 = IEEE float
    //}

    *p++ = (unsigned char)(nch >> 0);
    *p++ = (unsigned char)(nch >> 8);                   // Channels

    word32 = (unsigned int) (srate);
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);
    *p++ = (unsigned char)(word32 >> 16);
    *p++ = (unsigned char)(word32 >> 24);               // Sample frequency

    word32 *= Bytes * nch;
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);
    *p++ = (unsigned char)(word32 >> 16);
    *p++ = (unsigned char)(word32 >> 24);               // Bytes per second in the data stream

    word32 = Bytes * nch;
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);               // Bytes per sample time

    *p++ = (unsigned char)(bps >> 0);
    *p++ = (unsigned char)(bps >> 8);                   // Bits per single sample

    memcpy ( p, "data", 4 ); p += 4;

    word32 = (unsigned int)PCMdataLength;
    *p++ = (unsigned char)(word32 >>  0);
    *p++ = (unsigned char)(word32 >>  8);
    *p++ = (unsigned char)(word32 >> 16);
    *p++ = (unsigned char)(word32 >> 24);

    fwrite ( &Header, 1, sizeof(Header), stdout );
}

int main ( int argc, char **argv )
{
    int srate, bps, nch;
    char *buffer;

    if ( argc != 4 ) {
        printf ( "Usage: piperec <samplerate> <bps> <channels>\n" );
        printf ( "Output goes to stdout, hit any key to stop\n" );
        return 1;
    }

    srate = atoi ( argv[1] );
    bps   = atoi ( argv[2] );
    nch   = atoi ( argv[3] );

    buffer = (char *)malloc ( 10 * srate * nch * (bps+7)/8 );
    if ( !buffer ) {
        fprintf ( stderr, "Memory allocation failed.\n" );
        return 1;
    }

    if ( WindowsAudio_SetupInitInput (0, srate, bps, nch) != 0 ) {
        return 1;
    }

    _setmode ( _fileno(stdout), _O_BINARY );

    write_wav_header ( srate, bps, nch );

    while ( !_kbhit() ) {
        int bytes = WindowsAudio_record ( buffer );
        if ( bytes < 0 ) {
            fprintf ( stderr, "Error.\n" );
            break;
        }
        fwrite ( buffer, 1, bytes, stdout );
    }

    _getch();

    WindowsAudio_closeInput();

    return 0;
}
